--[[
Player.lua
Copyright (c) 2006 by Markus Inger, 2006 by Caeryn Dryad, 2007-2009 by Horacio Hoyos

This file is part of DHUD3.

    DHUD3 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DHUD3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DHUD3.  If not, see <http://www.gnu.org/licenses/>.
]]

local DHUD3 = LibStub("AceAddon-3.0"):GetAddon("DHUD3");
local L = LibStub("AceLocale-3.0"):GetLocale("DHUD3");
local DogTag = LibStub("LibDogTag-3.0");

local MODNAME = "Player";
local Player = DHUD3:NewModule(MODNAME, "AceEvent-3.0");

local unpack = unpack
local pairs = pairs
local type = type
local _G = _G
local string_find = string.find
local string_len = string.len
local string_sub = string.sub
local string_match = string.match
local UnitInParty = UnitInParty
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitPowerType = UnitPowerType
local UnitIsPlayer = UnitIsPlayer
local UnitIsPVP = UnitIsPVP
local UnitFactionGroup = UnitFactionGroup
local UnitIsPVPFreeForAll = UnitIsPVPFreeForAll
local UnitIsUnit = UnitIsUnit
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local GetRaidTargetIndex = GetRaidTargetIndex

local defaultfont = "Fonts/FRIZQT__.TTF"
local playerLayout = {
        pHealth = "e",
        pPower = "e",
        pIcons = "e",
        tHealth = "e",
        tPower = "e",
        tIcons = "e",
    }
--Fonts
local maxfont = 25
local minfont = 6    
--FIN

--Local Defaults
local db
local defaults = {
	profile = {
        blizPlayer = false,
		blizTarget = false,
        player = "c",
        combatIcon = true,
		pvpIcon = true,
        pvpIconTimer = true,
        swapPet = true, 
		restIcon= true,
		leaderIcon = true,
		lootIcon = true,
        pLayout = "HLPR",
        playerBarsText = {
			li = {},
			lm = {},
			ri = {},
			rm = {},
		},
        pBarHText = true,
		pBarPText = true,
		pHStyle = "[FractionalHP] ([PercentHP:Percent])",
		pCHStyle = "",
		pPStyle = "[FractionalMP] ([PercentMP:Percent])",
		pCPStyle = "",
		pBarTextSize = 10,
		pColors = {
	        ["0"] = {
	            full = { 0, 1, 1 },
	            med = { 0, 0, 1 },
	            low = { 1, 0, 1 }
	        },
	        ["1"] = {
	            full = { 1, 0, 0 },
	            med = { 1, 0.23, 0.23 },
	            low = { 1, 0.42, 0.42 }
	        },
	        ["3"] = { 
	            full = { 1, 1, 0 },
	            med = { 1, 1, 0.23 },
	            low = { 1, 1, 0.42 }
	        },
			["6"] = { 
	            full = { 0, 0.68, 0.87 },
	            med = { 0.23, 0.71, 0.84 },
	            low = { 0.47, 0.75, 0.81 }
	        },
			["7"] = {
	            full = { 0, 1, 0 },
	            med = { 1, 1, 0 },
	            low = { 1, 0, 0 }      
	        },
		},
		tBars = true,
		tNPC = true,
		tAuras = true,
		tTips = true,
		tSwap = true,
		tOwn = false,
		tBuffs = true,
		tDebuffs = true,
		tot = true,
		tName = true,
		tStyle = "[Level:DifficultyColor] [Classification] [Name:HostileColor] [isPet ? Creature:ClassColor:Bracket ! isPLayer ? Class:ClassColor:Bracket ! Faction:ClassColor:Bracket]",
		tCStyle = "",
		tNSize = 10,
		totName = true,
		totStyle = "[Level:DifficultyColor] [Classification] [Name:HostileColor] [isPet ? Creature:ClassColor:Bracket ! isPLayer ? Class:ClassColor:Bracket ! Faction:ClassColor:Bracket]",
		totCStyle = "",
		totNSize = 10,
		tPvpIcon = true,
		tEliteIcon = true,
		tBarHText = true,
		tBarPText = true,
		tHStyle = "[FractionalHP] ([PercentHP:Percent])",
		tCHStyle = "",
		tPStyle = "[FractionalMP] ([PercentMP:Percent])",
		tCPStyle = "",
		tBarTextSize = 10,
		tColors = {
			["0"] = {
	            full = { 0, 0.67, 0.67 },
	            med = { 0, 0, 0.67 },
	            low = { 0.67, 0, 0.67 }
	        },
			["1"] = { 
	            full = { 0.67, 0, 0 },
	            med = { 0.67, 0.15, 0.15 },
	            low = { 0.67, 0.28, 0.28 }
	        },
			["2"] = {
	            full = { 0.67, 0.27, 0 },
	            med = { 0.67, 0.27, 0 },
	            low = { 0.67, 0.27, 0 }
	        },
	        ["3"] = { 
	            full = { 0.67, 0.67, 0 },
	            med = { 0.67, 0.67, 0.15 },
	            low = { 0.67, 0.67, 0.28 }
	        },
			["6"] = { 
	            full = { 0, 0.43, 0.58 },
	            med = { 0.15, 0.48, 0.56 },
	            low = { 0.21, 0.50, 0.54 }
	        },
	        ["7"] = {
	            full = { 0, 0.67, 0 },
	            med = { 0.67, 0.67, 0 },
	            low = { 0.67, 0, 0 },
	        },
	        ["8"] = {
	            full = { 0.8, 0.8, 0.8 },
	            med = { 0.73, 0.73, 0.73 },
	            low = { 0.67, 0.67, 0.67 }
	        },
		},
    }
}
--FIN

-- Functions required by the configuration options table
local OptGetter, OptSetter, ColorGetter, ColorSetter, SelectGetter, SelectSetter, TBarsHidden, TNameHidden, TotNameHidden, BarTextHidden, TAurasEnable;
local layoutType, healthTags, powerTags, nameTags
do
	local mod = Player
	function OptGetter(info)
		local key = info[#info];
		return db[key];
	end

	function OptSetter(info, value)
		local key = info[#info];
		db[key] = value;
		mod:Refresh();
	end
    
    function SelectGetter (info, select)
        return db[select]
    end
    
    function SelectSetter (info, select, value) 
       db[select] = value
       mod:Refresh();
    end
    
    function ColorGetter (info)
        local key1, key2, key3 = info[#info-2], info[#info-1], info[#info]
		return db[key1][key2][key3][1], db[key1][key2][key3][2], db[key1][key2][key3][3]
	end
    
    function ColorSetter (info, r, g, b) 
        local key1, key2, key3 = info[#info-2], info[#info-1], info[#info]
        db[key1][key2][key3][1] = r
        db[key1][key2][key3][2] = g
        db[key1][key2][key3][3] = b
        mod:Refresh();
    end
    
    function TBarsHidden()
        return not db.tBars;
    end
    
    function TNameHidden()
        return not db.tName;
    end
    
    function TotNameHidden()
        return not db.totName;
    end
	
    function tDebuffEnable()
        return not db.tDebuff;
    end
	
    function TAurasEnable()
        return not db.tAuras;
    end
    
    function BarTextHidden(info)
        
        local key = info[#info];
        local who = string_sub(key, 1, 1);
        if string_find(key, "HStyle") then
            return not db[who.."BarHText"];
        elseif string_find(key, "PStyle") then
            return not db[who.."BarPText"];
        else
            return not (db[who.."BarHText"] or db[who.."BarPText"]);
        end
    end
    
    
    layoutType = {["c"] = L["Centered"], ["r"] = L["Right"], ["l"] = L["Left"]}
    healthTags = {
		["[FractionalHP] ([PercentHP:Percent])"] = "Current/Max (Percent%)",
		["[FractionalHP]"] = "Current/Max",
		["[HP] ([PercentHP:Percent])"] = "Current (Percent%)",
		["[PercentHP:Percent]"] = "Percent%",
		["[HP]"] = "Current"
	}
	powerTags = {
		["[FractionalMP] ([PercentMP:Percent])"] = "Current/Max (Percent%)",
		["[FractionalMP]"] = "Current/Max",
		["[MP] ([PercentMP:Percent])"] = "Current (Percent%)",
		["[PercentMP:Percent]"] = "Percent%", 
		["[MP]"] = "Current"
	}
    nameTags = {
	    ["[Level:DifficultyColor] [Classification] [Name:HostileColor] [isPet ? Creature:ClassColor:Bracket ! isPLayer ? Class:ClassColor:Bracket ! Faction:ClassColor:Bracket]"] = "<level><elite> <name> [ <class> ]",
	    ["[Level:DifficultyColor] [Classification] [Name:HostileColor]"] = "<level><elite> <name>",
	    ["[Name:HostileColor] [isPet ? Creature:ClassColor:Bracket ! isPLayer ? Class:ClassColor:Bracket ! Faction:ClassColor:Bracket]"] = "<name> [ <class> ]",
	    ["[Level:DifficultyColor] [Name:HostileColor]"] = "<level> <name>",
    }
end

--Local Options
local options
local function GetOptions()
	if not options then
        --Options table
    	options = {
    		type = "group",
			name = L["The player module manages the two main bars (player and target) functionality."],
			arg = MODNAME,
            order = 2,
			get = OptGetter,
			set = OptSetter,
            childGroups = "tab",
    		args = {
                enabled = {
                    order = 1,
                    type = "toggle",
                    name = L["Enable Player Module"],
                    get = function() return DHUD3:GetModuleEnabled(MODNAME) end,
                    set = function(info, value) DHUD3:SetModuleEnabled(MODNAME, value) end,
                },
                player = {
                    type = "group",
                    order = 2,	
                    name = L["Player Options"],
                    desc = L["Configure the players's layout options"],
					disabled  = function() return not DHUD3:GetModuleEnabled(MODNAME); end,
                    args = {
                        pLayout = {
                            type = 'group',
                            order = 1,
                            name = L["Layout"],
                            args = {
                                player = {
                                    type = 'select',
                                    order = 1,
                                    name = L["Player/Target Bars"],
                                    desc = L["Place the Player's bars. Target's bars will be set accordingly"],
                                    set = function (info, value) 
                                            db[info[#info]] = value
                                            if string_match(value, "c") then
                                                db.pLayout = "HLPR";
                                                db.tLayout = "HLPR";
                                            else
                                                db.pLayout = "HIPM";
                                                db.tLayout = "HIPM";
                                            end
                                            Player:Refresh();
                                        end,
                                    values = layoutType,
                                },
                                pLayout = {
                                    type = 'select',
                                    order = 2,
                                    name = L["Bar Layout"],
                                    values = function ()
                                            if string_match(db.player, "c") then
                                                return { ["HLPR"] = L["Health Left/Power Right"], ["HRPL"] = L["Health Right/Power Left"] }
                                            else
                                                return { ["HIPM"] = L["Health Inner/Power Outter"], ["HMPI"] = L["Health Outter/Power Inner"] }
                                            end
                                        end,
                                },
                                swapPet = {
                                    type = 'toggle',
                                    name = L["Bar Swap"],
                                    desc = L["Swap player's status for pet's when using vehicles"],
                                    order = 3,
                                },
                                icons = {
                                    type = 'header',
                                    name = L["Status Icons"],
                                    order = 4,
                                },
                                combatIcon = {
                                    type = 'toggle',
                                    order = 5,
                                    name = L["In Combat"],
                                    desc ="Show/Hide",
                                },
                                restIcon = {
                                    type = 'toggle',
                                    order = 6,
                                    name = L["Resting"],
                                    desc = L["Show/Hide"],
                                },
                                leaderIcon = {
                                    type = 'toggle',
                                    order = 7,
                                    name = L["Party Leader"],
                                    desc = L["Show/Hide"],
                                },
                                lootIcon = {
                                    type = 'toggle',
                                    order = 8,
                                    name = L["Master Looter"],
                                    desc = L["Show/Hide"],
                                },
                                pvpIcon = {
                                    type = 'toggle',
                                    order = 9,
                                    name = L["PVP Flag"],
                                    desc = L["Show/Hide"],
                                },
                                pvpTimer = {
                                    type = 'header',
                                    name = L["PvP Timer"],
                                    order = 10,
                                },
                                pvpIconTimer = {
                                    type = 'toggle',
                                    order = 11,
                                    name = L["PvP Timer"],
                                    desc = L["Show/Hide"],
                                    disabled  = function() return not db.pvpIcon; end,
                                },
                            },
                        },
                        barText = {
                            type = 'group',
                            order = 2,
                            name = L["Bar's Text"],
                            desc = L["Configure Health and Power (mana/rage/energy/runic) text options"],
                            args = {
                                pBarText = {
                                    type = 'multiselect',
                                    order = 1,
                                    name = L["Show Bar Values"],
                                    values = { ["pBarHText"] = L["Health"], ["pBarPText"] = L["Power"] },
                                    get = SelectGetter,
                                    set = SelectSetter,
                                },
                                pHStyle = {
                                    type = 'select',
                                    order = 2,
                                    name = L["Health Style"],
                                    hidden = BarTextHidden,
                                    values = healthTags
                                    },
                                pCHStyle = {
                                    type = 'input',
                                    order = 3,
                                    name = L["Custom Health Style"],
                                    desc = L["If set this style will override Health Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
                                    hidden = BarTextHidden,
                                    usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],							
                                },
                                pPStyle = {
                                    type = 'select',
                                    order = 4,
                                    name = L["Power Style"],
                                    hidden = BarTextHidden,
                                    values = powerTags
                                    },
                                pCPStyle = {
                                    type = 'input',
                                    order = 5,
                                    name = L["Custom Power Style"],
                                    desc = L["If set this style will override Power Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
                                    hidden = BarTextHidden,
                                    usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],								
                                },
                                pBarTextSize = {
                                    type = 'range',
                                    order = 6,
                                    name = L["Font size"],
                                    hidden = BarTextHidden,
                                    min = minfont,
                                    max = maxfont,
                                    step = 1,
                                },
                            },
                        },	
                        pColors = {
                            type = 'group',
                            order = 3,
                            name = L["Colors"],
                            get = ColorGetter,
                            set = ColorSetter,
                            args = {
                                ["7"] = {
                                    type = 'group',
                                    name = L["Health"],
                                    order = 1,
                                    guiInline  = true,
                                    args = {
                                        full = {
                                            type = 'color',
                                            name = L["Full"],
                                            order = 1,
                                        },
                                        med = {
                                            type = 'color',
                                            name = L["Med"],
                                            order = 2,
                                        },
                                        low = {
                                            type = 'color',
                                            name = L["Low"],
                                            order = 3,
                                        },
                                    },
                                },
                                ["0"] = {
                                    type = 'group',
                                    name = L["Mana"],
                                    guiInline  = true,
                                    args = {
                                        full = {
                                            type = 'color',
                                            name = L["Full"],
                                            order = 1,
                                        },
                                        med = {
                                            type = 'color',
                                            name = L["Med"],
                                            order = 2,
                                        },
                                        low = {
                                            type = 'color',
                                            name = L["Low"],
                                            order = 3,
                                        },
                                    },
                                },
                                ["1"] = {
                                    type = 'group',
                                    name = L["Rage"],
                                    guiInline  = true,
                                    args = {
                                        full = {
                                            type = 'color',
                                            name = L["Full"],
                                            order = 1,
                                        },
                                        med = {
                                            type = 'color',
                                            name = L["Med"],
                                            order = 2,
                                        },
                                        low = {
                                            type = 'color',
                                            name = L["Low"],
                                            order = 3,
                                        },
                                    },
                                },
                                ["3"] = { 
                                    type = 'group',
                                    name = L["Energy"],
                                    guiInline  = true,
                                    args = {
                                        full = {
                                            type = 'color',
                                            name = L["Full"],
                                            order = 1,
                                        },
                                        med = {
                                            type = 'color',
                                            name = L["Med"],
                                            order = 2,
                                        },
                                        low = {
                                            type = 'color',
                                            name = L["Low"],
                                            order = 3,
                                        },
                                    },
                                },
                                ["6"] = { 
                                    type = 'group',
                                    name = L["Runic Power"],
                                    guiInline  = true,
                                    args = {
                                        full = {
                                            type = 'color',
                                            name = L["Full"],
                                            order = 1,
                                        },
                                        med = {
                                            type = 'color',
                                            name = L["Med"],
                                            order = 2,
                                        },
                                        low = {
                                            type = 'color',
                                            name = L["Low"],
                                            order = 3,
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
                target = {
                    type = 'group',
    				order = 3,	
                    name = L["Target Options"],
                    desc = L["Configure the target's layout options"],
					disabled  = function() return not DHUD3:GetModuleEnabled(MODNAME); end,
                    args = {
                        tLayout = {
                            type = 'group',
                            order = 1,
                            name = L["Layout"],
                            args = {
                                tBars = {
                                    type = 'toggle',
                                    order = 1,
                                    name = L["Target Bars"],
                                    desc = L["Show/Hide"],
                                },
                                tNPC = {
                                    type = 'toggle',
                                    order = 2,
                                    name = L["When NPC"],
                                    desc = L["Show/Hide"],
                                    hidden = TBarsHidden,
                                },
                                icons = {
                                    type = 'header',
                                    name = L["Status Icons"],
                                    order = 3,
                                    hidden = TBarsHidden,
                                },
                                tPvpIcon = {
                                    type = 'toggle',
                                    order = 4,
                                    name = L["Show PVP Flag / Raid Icon"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                },
                                tEliteIcon = {
                                    type = 'toggle',
                                    order = 5,
                                    name = L["Show Elite"],
                                    desc = L["Show/Hide"],
                                    hidden = TBarsHidden,
                                },
                                auras = {
                                    type = 'header',
                                    name = L["Auras"],
                                    order = 6,
                                    hidden = TNameHidden,
                                },
                                tAuras = {
                                    type = 'toggle',
                                    order = 7,
                                    name = L["Target Auras"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                },
								tBuffs = {
                                    type = 'toggle',
                                    order = 9,
                                    name = L["Show Target Buffs"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                    disabled = TAurasEnable,
                                },
								tDebuffs = {
                                    type = 'toggle',
                                    order = 10,
                                    name = L["Show Target Debuffs"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
									disabled = TAurasEnable,
                                },
								tOwn = {
                                    type = 'toggle',
                                    order = 8,
                                    name = L["Show only my Debuffs"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                    disabled = tDebuffEnable,
                                },
                                tSwap = {
                                    type = 'toggle',
                                    order = 11,
                                    name = L["Swap Target Auras"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                    disabled = TAurasEnable,
                                },
                                tTips = {
                                    type = 'toggle',
                                    order = 12,
                                    name = L["Show Aura Tips"],
                                    desc = L["Show/Hide"],
                                    hidden = TNameHidden,
                                    disabled = TAurasEnable,
                                },
                                tName = {
                                    type = 'group',
            						order = 13,
                                    name = L["Target Name Plate"],
                                    args = {
            							tName = {
            								type = 'toggle',
            								order = 1,
            								name = L["Show"],
            								desc = L["Show target Name plate"],
            							},
            							tStyle = {
            								type = 'select',
            								order = 2,
            								name = L["Name Style"],
            								hidden = TNameHidden,
            								values = nameTags
            							},
            							tCStyle = {
            								type = 'input',
            								order = 3,
            								name = L["Custom Name Style"],
            								desc = L["If set this style will override Name Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
            								hidden = TNameHidden,
            								usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],							
            							},
                                        tNSize = {
                                            type = 'range',
            								order = 4,
                                            name = L["Size"],
            								hidden = TNameHidden,
                                            min = minfont,
                                            max = maxfont,
                                            step = 1,
                                        },
                                    },
                                },
                                totName = {
                                    type = 'group',
            						order = 14,
                                    name = L["Target Target Name Plate"],
                                    args = {
            							totName = {
            								type = 'toggle',
            								order = 1,
            								name = L["Show"],
            								desc = L["Show target of target Name plate"],
            							},
            							totStyle = {
            								type = 'select',
            								order = 2,
            								name = L["Name Style"],
            								hidden = TotNameHidden,
            								values = nameTags
            							},
            							totCStyle = {
            								type = 'input',
            								order = 3,
            								name = L["Custom Name Style"],
            								desc = L["If set this style will override Name Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
            								hidden = TotNameHidden,
            								usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],								
            							},
                                        totNSize = {
                                            type = 'range',
            								order = 4,
                                            name = L["Size"],
            								hidden = TotNameHidden,
                                            min = minfont,
                                            max = maxfont,
                                            step = 1,
                                        },
                                    },
                                },
                            },
                        },
                        barText = {
    						type = 'group',
    						order = 5,
    						name = L["Bar's Text"],
                            desc = L["Configure Health and Power (mana/rage/energy/runic) text options"],
    						args = {
    							tBarText = {
    								type = 'multiselect',
    								order = 1,
    								name = L["Show Bar Values"],
    								values = { ["tBarHText"] = L["Health"], ["tBarPText"] = L["Power"] },
                                    get = SelectGetter,
                                    set = SelectSetter,
    							},
    							tHStyle = {
    								type = 'select',
    								order = 2,
    								name = L["Health Style"],
    								hidden = BarTextHidden,
    								values = healthTags
    								},
    							tCHStyle = {
    								type = 'input',
    								order = 3,
    								name = L["Custom Health Style"],
    								desc = L["If set this style will override Health Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
    								hidden = BarTextHidden,
    								usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],							
    							},
    							tPStyle = {
    								type = 'select',
    								order = 4,
    								name = L["Power Style"],
    								hidden = BarTextHidden,
    								values = powerTags
    								},
    							tCPStyle = {
    								type = 'input',
    								order = 5,
    								name = L["Custom Power Style"],
    								desc = L["If set this style will override Health Style value. Use LibDogTag-3.o Tags. (Note: Leave/set blank to use Heatlh Style setting.)"],
    								hidden = BarTextHidden,
    								usage = L["Enter /dog for in-game LibDogTag-3.0 Tag documentation."],								
    							},
    							tBarTextSize = {
    		                        type = 'range',
    								order = 6,
    		                        name = L["Font size"],
    								hidden = BarTextHidden,
    		                        min = minfont,
    		                        max = maxfont,
    		                        step = 1,
    		                        },
    						},
    					},	
                        tColors = {
    						type = 'group',
    						order = 6,
    						name = L["Bar Colors"],
                            get = ColorGetter,
                            set = ColorSetter,
    						hidden = TBarsHidden,
    						args = {
    		                    ["7"] = {
    		                        type = 'group',
    								guiInline  = true,
    		                        name = L["Health"],
    								order = 1,
    		                        args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
    		                    },
    		                    ["0"] = {
    		                        type = 'group',
    								guiInline  = true,
    		                        name = L["Mana"],
    		                        args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
    		                    },
    		                    ["1"] = {
    		                        type = 'group',
    								guiInline  = true,
    		                        name = L["Rage"],
    		                        args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
    		                    },
    		                    ["3"] = { 
    		                        type = 'group',
    								guiInline  = true,
    		                        name = L["Energy"],
    		                        args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
    		                    },
    							["6"] = { 
    		                        type = 'group',
    								guiInline  = true,
    		                        name = L["Runic Power"],
    		                        args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
                                },
    		                    ["2"] = {
                                    type = 'group',
    								guiInline  = true,
                                    name = L["Focus (Pet)"],
                                    args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
                                },
                                ["8"] = {
                                    type = 'group',
    								guiInline  = true,
                                    name = L["Tapped"],
                                    args = {
    		                            full = {
    		                                type = 'color',
    		                                name = L["Full"],
    										order = 1,
    		                            },
    		                            med = {
    		                                type = 'color',
    		                                name = L["Med"],
    										order = 2,
    		                            },
    		                            low = {
    		                                type = 'color',
    		                                name = L["Low"],
    										order = 3,
    		                            },
    		                        },
                                },
    						},
    					},
    				},
                },
            }
        }
    end
    
    return options;
end

-- Local Functions
local blizzFrames, ConfigBars, PlaceBarTexts, UpdateTargetAuras, HideTargetAuras, UpdateTargetElite, TargetDropDownInit, PlayerDropDownInit, HideTargetAuras, SwapVehicle, UpdatePlayerPvP, OnUnitHealth, OnUnitPower, OnUpdateBarPower, OnBarShow, OnBarHide
do
	--[[[
	    blizzFrames Manage the state of blizzard frames
		--]]
	function BlizzFrames(showPlayer, showTarget)
	    
	    if showPlayer then
	        PlayerFrame:Show();
	    else
	        PlayerFrame:Hide();
	    end    

	    if showTarget then
	        TargetFrame:TargetFrame_Update (_G["TargetFrame"]);
	    else
	        TargetofTargetFrame:Hide();
	        ComboFrame:Hide();
	        TargetFrame:Hide();
	    end
	end
	
    -- Asing each bar the their respective tracking stat accordingly to the saved profile
    function ConfigBars()
        --DHUD3:Debug("ConfigBars");
        -- Player Bars
    	if string_match(db.player, "c") then
    		if string_match(db.pLayout, "HLPR") then
                playerLayout.pHealth = "li";
                playerLayout.pPower = "ri";
                Player.auraSide = "r";
                playerLayout.tHealth = "lm";
                playerLayout.tPower = "rm";
                playerLayout.pIcons = "l";
                playerLayout.tIcons = "r";
    		else
                playerLayout.pHealth = "ri";
                playerLayout.pPower = "li";
                Player.auraSide = "l";
                playerLayout.tHealth = "rm";
                playerLayout.tPower = "lm";
                playerLayout.pIcons = "r";
                playerLayout.tIcons = "l";
    		end
    	elseif string_match(db.player, "l") then
            playerLayout.pIcons = "l";
            Player.auraSide = "l";
            playerLayout.tIcons = "r";
    	    if string_match(db.pLayout, "HIPM") then
                playerLayout.pHealth = "li";
                playerLayout.pPower = "lm";
                playerLayout.tHealth = "ri";
                playerLayout.tPower = "rm";
    		else
                playerLayout.pHealth = "lm";
                playerLayout.pPower = "li";
                playerLayout.tHealth = "rm";
                playerLayout.tPower = "ri";
            end
    	else
    		playerLayout.pIcons = "r";
            Player.auraSide = "r";
            playerLayout.tIcons = "l";
    	    if string_match(db.pLayout, "HIPM") then
                playerLayout.pHealth = "ri";
                playerLayout.pPower = "rm";
                playerLayout.tHealth = "li";
                playerLayout.tPower = "lm";
    		else
                playerLayout.pHealth = "rm";
                playerLayout.pPower = "ri";
                playerLayout.tHealth = "lm";
                playerLayout.tPower = "li";
            end
        end
    	
    	--Bars Text Tag
    	if db.pBarHText then
            _G["DHUD3_"..playerLayout.pHealth.."_Text"]:Show();
    		local tagCode = (string_len(db.pCHStyle) > 0) and db.pCHStyle or db.pHStyle;
    		DogTag:AddFontString(_G["DHUD3_"..playerLayout.pHealth.."_Text_Text"], _G["DHUD3_"..playerLayout.pHealth.."_Text_Text"], tagCode, "Unit", { unit = "player" } );
        else
            DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.pHealth.."_Text_Text"]);
            _G["DHUD3_"..playerLayout.pHealth.."_Text"]:Hide();
    	end
    	if db.pBarPText then
            _G["DHUD3_"..playerLayout.pPower.."_Text"]:Show();
    		local tagCode = (string_len(db.pCPStyle) > 0) and db.pCPStyle or db.pPStyle;
    		DogTag:AddFontString(_G["DHUD3_"..playerLayout.pPower.."_Text_Text"], _G["DHUD3_"..playerLayout.pPower.."_Text_Text"], tagCode, "Unit", { unit = "player" });
        else
            DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.pPower.."_Text_Text"]);
            _G["DHUD3_"..playerLayout.pPower.."_Text"]:Hide();
    	end
        
        
    	if db.tBarHText then
            _G["DHUD3_"..playerLayout.tHealth.."_Text"]:Show();
    		local tagCode = (string_len(db.tCHStyle) > 0) and db.tCHStyle or db.tHStyle;
    		DogTag:AddFontString(_G["DHUD3_"..playerLayout.tHealth.."_Text_Text"], _G["DHUD3_"..playerLayout.tHealth.."_Text_Text"], tagCode, "Unit", { unit = "target" } );
        else
            DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.tHealth.."_Text_Text"]);
            _G["DHUD3_"..playerLayout.tHealth.."_Text"]:Hide();
    	end
    	if db.tBarPText then
            _G["DHUD3_"..playerLayout.tPower.."_Text"]:Show();
    		local tagCode = (string_len(db.tCPStyle) > 0) and db.tCPStyle or db.tPStyle;
    		DogTag:AddFontString(_G["DHUD3_"..playerLayout.tPower.."_Text_Text"], _G["DHUD3_"..playerLayout.tPower.."_Text_Text"], tagCode, "Unit", { unit = "target" });
        else
            DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.tPower.."_Text_Text"]);
            _G["DHUD3_"..playerLayout.tPower.."_Text"]:Hide();
    	end
    end

    -- Place the bar's text if player moved them
    function PlaceBarTexts()
    	
    	for k,v in pairs(db.playerBarsText) do
    		--DHUD3:Debug("PlaceBarTexts", k, v.moved, v.x, v.y)
    		local barText = DHUD3.layout.barText
    		w = barText[k]
    		local point, parent, relative, x, y = unpack(w)
    		ref = _G["DHUD3_"..k.."_Text"]
    		ref:ClearAllPoints()
    		if v.moved then
    			ref:SetPoint(point, _G["DHUD3_"..parent], relative, v.x, v.y)
    		else
    			ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y)
    		end
    	end
    end
    
    -- Update target auras
    function UpdateTargetAuras()
	
        local buffBorder, buffText, buffFrame;
        local debuffBorder, debuffText, debuffFrame;
        
    	if not db.tSwap then
    		buffFrame   = "DHUD3_TLA";
    		debuffFrame = "DHUD3_TRA";
    	else
    		buffFrame   = "DHUD3_TRA";    
    		debuffFrame = "DHUD3_TLA";
    	end
        
    	-- Buffs
		if db.tBuffs then
	    	for i = 1, 40 do
	    		local button = _G[buffFrame..i]
	    		local name, _, texture, count = UnitBuff("target", i);
	    		if ( name ) then
	                button.hasDebuff = false;
	                button.id = i;
	    			button:SetNormalTexture(texture);
	    			_G[buffFrame..i.."_Border"]:Hide();
	    			buffText = _G[buffFrame..i.."_Text"];
	    			if ( count > 1 ) then
	    				buffText:SetText(count);
	    			else
	    				buffText:SetText("");
	    			end
	    			button:Show();
	                button:EnableMouse(db.tTips);
	    		else    
	    			button:Hide();
	    		end
	    	end
		end
    	
    	-- DeBuffs
		if db.tDebuffs then
	    	for i = 1, 40 do
	    		local button = _G[debuffFrame..i]
	    		local name, _, texture, count = UnitDebuff("target", i)
	    		if ( name ) then
	                button.hasDebuff = true;
	                button.id = i;
	    			button:SetNormalTexture(texture)
	    			_G[debuffFrame..i.."_Border"]:Show()
	    			debuffText = _G[debuffFrame..i.."_Text"]
	    			if ( count > 1 ) then
	    				debuffText:SetText(count)
	                else
	    				debuffText:SetText("")
	    			end
	    			button:Show()
	                button:EnableMouse(db.tTips);
	    		else
	    			button:Hide()
	    		end
	        end
		end
    end
	
	-- Hide all target auras
	function HideTargetAuras()
		for i = 1, 40 do
			_G["DHUD3_TLA"..i]:Hide();
			_G["DHUD3_TRA"..i]:Hide();
		end		
	end

	-- Show targets Elite icon
    function UpdateTargetElite(show)
    	local elite = UnitClassification("target")
    	local texture = ""
        if elite and show then
            --local images = self.layout.icons
    		if elite == "worldboss" then
    			texture = "Interface\\AddOns\\DHUD3\\textures\\elite"
    		elseif elite == "rareelite" then
    			texture = "Interface\\AddOns\\DHUD3\\textures\\rareelite"
    		elseif elite == "rare" then
    			texture = "Interface\\AddOns\\DHUD3\\textures\\rare"
    		elseif elite == "elite" then
    			texture = "Interface\\AddOns\\DHUD3\\textures\\elite"
    		elseif elite == "normal" then
    			texture = ""
    		elseif elite == "trivial" then
    			texture = ""
    		else
    			texture = ""
    		end
    		_G["DHUD3_"..playerLayout.tIcons.."telite_Texture"]:SetTexture(texture)
    		_G["DHUD3_"..playerLayout.tIcons.."telite"]:Show()
    	else
    		_G["DHUD3_"..playerLayout.tIcons.."telite"]:Hide()
    	end
    end
    
    --Initialize the target dropdown menu, taken from blizz code.
    function TargetDropDownInit(self)
        local menu;
    	local name;
    	local id = nil;
    	if ( UnitIsUnit("target", "player") ) then
    		menu = "SELF";
    	elseif ( UnitIsUnit("target", "vehicle") ) then
    		-- NOTE: vehicle check must come before pet check for accuracy's sake because
    		-- a vehicle may also be considered your pet
    		menu = "VEHICLE";
    	elseif ( UnitIsUnit("target", "pet") ) then
    		menu = "PET";
    	elseif ( UnitIsPlayer("target") ) then
    		id = UnitInRaid("target")
    		if ( id ) then
    			menu = "RAID_PLAYER";
    			name = GetRaidRosterInfo(id +1);
    		elseif ( UnitInParty("target") ) then
    			menu = "PARTY";
    		else
    			menu = "PLAYER";
    		end
    	else
    		menu = "RAID_TARGET_ICON";
    		name = RAID_TARGET_ICON;
    	end
    	if ( menu ) then
    		UnitPopup_ShowMenu(self, menu, "target", name, id);
    	end
    end

    -- Initialize the target dropdown menu, taken from blizz code.
    function PlayerDropDownInit()
        UnitPopup_ShowMenu( DHUD3_playerDropDown, "SELF", "player" );
    end
	
	function SwapVehicle(swap)
	
		--DHUD3:Debug("SwapVehicle", swap);
	    local maxVal, curFrac,needH, needP;
	    local color = {r=0,g=0,b=0,hex=""};
	    if swap then
	        who = "vehicle";
	    else
	        who = "player";
	    end
	    -- Health
		maxVal, curFrac, needH = DHUD3:GetUnitHealth(who);
		local colors = db.pColors[tostring(7)];
		color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
		DHUD3:UpdateBar(playerLayout.pHealth, maxVal, curFrac, color);
		local bar = _G["DHUD3_"..playerLayout.pHealth]
		bar.who = who;
		bar.colors = colors;
		bar.swap = swap;
		--Power
		maxVal, curFrac, needP = DHUD3:GetUnitPower(who);
		colors = db.pColors[tostring(UnitPowerType(who))];
		color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
		DHUD3:UpdateBar(playerLayout.pPower, maxVal, curFrac, color);    
		bar = _G["DHUD3_"..playerLayout.pPower]
		bar.who = who;
		bar.maxVal = maxVal;
		bar.curFrac = curFrac;
		if ( maxVal ~= 0 ) then 
			bar:SetScript("OnUpdate", Player.OnUpdateBarPower);
			local powerType = UnitPowerType(who);
			bar.display = powerType;
			bar.colors = colors;
			if ( powerType == 0 ) then
				bar:RegisterEvent("UNIT_MAXMANA");
			elseif ( powerType == 1 ) then
				bar:RegisterEvent("UNIT_MAXRAGE");
			elseif ( powerType == 3 ) then
				bar:RegisterEvent("UNIT_MAXENERGY");
			elseif ( powerType == 6 ) then
				bar:RegisterEvent("UNIT_MAXRUNIC_POWER");
			end
		else
			bar:SetScript("OnUpdate", nil);
		end
		--Bars Text Tag
		if db.pBarHText then
			DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.pHealth.."_Text_Text"]);
			local tagCode = (string_len(db.pCHStyle) > 0) and db.pCHStyle or db.pHStyle;
			DogTag:AddFontString(_G["DHUD3_"..playerLayout.pHealth.."_Text_Text"], _G["DHUD3_"..playerLayout.pHealth.."_Text_Text"], tagCode, "Unit", { unit = who} );
		end
		if db.pBarPText then
			DogTag:RemoveFontString(_G["DHUD3_"..playerLayout.pPower.."_Text_Text"]);
			local tagCode = (string_len(db.pCPStyle) > 0) and db.pCPStyle or db.pPStyle;
			DogTag:AddFontString(_G["DHUD3_"..playerLayout.pPower.."_Text_Text"], _G["DHUD3_"..playerLayout.pPower.."_Text_Text"], tagCode, "Unit", { unit = who });
		end
	end
	
	function UpdatePlayerPvP(show, side)
	    if ( db.pvpIcon ) then
	        local _, _, _, _, _, _, _, x0, x1, y0, y1, path = unpack(DHUD3.layout.icons.target);
	        if ( UnitIsPVPFreeForAll("player") ) then
	            path = path[3];
	        elseif ( UnitIsPVP("player") ) then 
	            local faction = UnitFactionGroup("player");
	            if ( faction == "Horde" ) then
	                path = path[2];
	            else
	                path = path[1];
	            end
	        else
	            _G["DHUD3_"..playerLayout.pIcons.."pvp"]:Hide();
	            return;
	        end
	        local texture = _G["DHUD3_"..playerLayout.pIcons.."pvp_Texture"];
	        texture:SetTexture(path);
	        texture:SetTexCoord(x0,x1,y0,y1);
	        _G["DHUD3_"..playerLayout.pIcons.."pvp"]:Show();
	    else
	        _G["DHUD3_"..playerLayout.pIcons.."pvp"]:Hide();
	    end
	end
	
	function OnUnitHealth(this, event, who)
		--DHUD3:Debug("OnUnitHealth",this, event, who);
	    if who == this.who then
	        local maxVal, curFrac, need
	        local color = {r=0,g=0,b=0,hex=""}
	        maxVal, curFrac, need = DHUD3:GetUnitHealth(who)
	        color.r, color.g, color.b, color.hex = DHUD3:Colorize(this.colors, curFrac);
	        DHUD3:UpdateBar(this.bar, maxVal, curFrac, color);
	        if ( need ) then 
	            DHUD3:UpdateAlpha(1, true);
			else
				DHUD3:UpdateAlpha(1, false);
	        end
	    end
	end

	function OnUnitPower(this, event, who)
		    
		if ( event == "UNIT_DISPLAYPOWER" ) then
			if who == this.who then
				local powerType = UnitPowerType(who);
				this.display = powerType;
				this.colors = db.pColors[tostring(powerType)];
				this.maxVal = UnitPowerMax(this.who, powerType);
			end
		else
			if who == this.who then
				this.maxVal = UnitPowerMax(this.who, this.display);
			end
		end
	end

	function OnUpdateBarPower(event, elapsed)
		
	    local color = {r=0,g=0,b=0,hex=""};
		local need;
	    local curFrac = UnitPower(this.who, this.display) / this.maxVal;
		
	    if ( this.display == 1 or this.display == 6 ) then
			if curFrac ~= 0 then 
				need = true;
			end
		else
			if curFrac ~= 1 then 
				need = true;
			end
		end
		this.curVal = curVal;
		color.r, color.g, color.b, color.hex = DHUD3:Colorize(this.colors, curFrac);
		DHUD3:UpdateBar(this.bar, maxVal, curFrac, color);
	    if ( need ) then 
			DHUD3:UpdateAlpha(1, true);
		else
			DHUD3:UpdateAlpha(1, false);
		end
end

	function OnBarShow(this, event)
		--DHUD3:Debug("OnBarShow", this, event);
		DHUD3:SendMessage("DHUD3_BAR_SHOW", this.bar);
	end
	
	function OnBarHide(this, event)
		--DHUD3:Debug("OnBarHide", this, event);
		DHUD3:SendMessage("DHUD3_BAR_HIDE", this.bar);
	end
end

function Player:OnInitialize()
	self.db = DHUD3.db:RegisterNamespace(MODNAME, defaults);
	db = self.db.profile;
	
	self:SetEnabledState(DHUD3:GetModuleEnabled(MODNAME));
	DHUD3:RegisterModuleOptions(MODNAME, GetOptions, L["Player"]);
    
    
    local icons = DHUD3.layout.icons;
    local bars = DHUD3.layout.playerBars;
    local barText = DHUD3.layout.barText;
    local texts = DHUD3.layout.texts;
    local tRA = DHUD3.layout.targetRightAuras;
    local tLA = DHUD3.layout.targetLeftAuras;
    local point, parent, relative, x, y, width, height;
    
    -- Bars
    for k, v in pairs(bars) do
        point, parent, relative, position, texture, border = unpack(v);
        x0, x1, y0, y1 = unpack(DHUD3.layout[position]);
        ref = CreateFrame("Button", "DHUD3_"..k, _G["DHUD3_"..parent]);
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, 0, 0);
        ref:SetHeight(256);
        ref:SetWidth(128);
        bgt = ref:CreateTexture("DHUD3_"..k.."_Texture", "ARTWORK");
        bgt:SetTexture(texture);
        bgt:SetPoint(point, ref, relative, 0, 0);
        bgt:SetHeight(256);
        bgt:SetWidth(128);
        bgt:SetTexCoord(x0, x1, y0, y1);
        bgt:SetBlendMode("BLEND");
        bgt = ref:CreateTexture("DHUD3_"..k.."_Border", "BACKGROUND");
        bgt:SetTexture(border);
        bgt:SetPoint(point, ref, relative, 0, 0);
        bgt:SetHeight(256);
        bgt:SetWidth(128);
        bgt:SetTexCoord(x0, x1, y0, y1);
        bgt:SetBlendMode("BLEND");
        ref:EnableMouse(false);
        ref:Hide();
		ref:SetScript("OnShow", OnBarShow);
		ref:SetScript("OnHide", OnBarHide);
    end

    -- Bar's Text
    for k, v in pairs(barText) do
        point, parent, relative, x, y, width, height, justify = unpack(v);
        ref = CreateFrame("Button", "DHUD3_"..k.."_Text", _G["DHUD3_"..parent]);
        ref:SetWidth(width);
        ref:SetHeight(height);
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y);
        local font = ref:CreateFontString("DHUD3_"..k.."_Text_Text", "OVERLAY");
        font:SetFontObject(GameFontHighlightSmall);
        font:ClearAllPoints();
        font:SetPoint(justify, ref, justify, 0, 0);
		font:SetText(k.."2.5 Lets try the aligment");
        font:SetWidth(font:GetStringWidth() * 1.1);
        font:SetJustifyH(justify);
        font:Show();
        ref.isMoving = false;
		ref.dropTimer = 0;
		ref:RegisterForDrag("LeftButton");
		ref:SetMovable(true);
		ref:SetScript("OnDragStart", self.MoveFrame);
		ref:SetScript("OnDragStop", self.DropFrame);
        ref:EnableMouse(true);
        ref:Hide();
    end

    -- Text
    for k, v in pairs(texts) do
        point, parent, relative, x, y, width, height = unpack(v);
        ref = CreateFrame("Button", "DHUD3_"..k.."_Text", _G["DHUD3_"..parent], "SecureActionButtonTemplate");
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y);
        local font = ref:CreateFontString("DHUD3_"..k.."_Text_Text", "OVERLAY");
        font:SetFontObject(GameFontHighlightSmall);
        font:ClearAllPoints();
        font:SetPoint("CENTER", ref, "CENTER", 0, 0);
        font:SetText(k);
        font:SetJustifyH("CENTER");
		font:SetJustifyV("BOTTOM");
        font:SetWidth(330);
        font:Show();
        ref:SetHeight(height);
        ref:SetWidth(350);
        ref:Hide();
    end
    
    -- Icons
    for k, v in pairs(icons) do
        point, parent, relative, x, y, width, height, x0, x1, y0, y1, path = unpack(v);
        if string_find(k, "PvP") then         
            local faction = UnitFactionGroup("player");
            if faction == "Horde" then
                path = path[2];
            else
                path = path[1];
            end
        end
        ref = CreateFrame("Frame", "DHUD3_"..k, _G["DHUD3_"..parent]);
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y);
        ref:SetHeight(height);
        ref:SetWidth(width);
        local bgt = ref:CreateTexture("DHUD3_"..k.."_Texture","OVERLAY");
        bgt:SetTexture(path);
        bgt:ClearAllPoints();
        bgt:SetPoint("TOPLEFT", ref , "TOPLEFT", 0, 0);
        bgt:SetPoint("BOTTOMRIGHT", ref , "BOTTOMRIGHT", 0, 0);
        bgt:SetTexCoord(x0,x1,y0,y1);
		if k == "rpvp" or k == "lpvp" then
			ref:SetScript("OnEnter", function(this)
                if not this:IsVisible() then
                    return;
                end
				local left = GetPVPTimer();
                local text;
                if ( left == 301000 ) then
                    text = "PvP On";
                else
                    text = DHUD3:formatTime(left/1000);
                end
                if playerLayout.pIcons == "r" then 
                    GameTooltip:SetOwner(this, "ANCHOR_TOPLEFT");
                else
                    GameTooltip:SetOwner(this, "ANCHOR_TOPRIGHT");
                end
                GameTooltip:SetText(text, 1, 0, 0);
            end )
			ref:SetScript("OnLeave", function() 
                GameTooltip:Hide();
            end )        
			ref:EnableMouse(true);
		end
        ref:Hide();
    end
    
    -- Target Auras
    for i = 1, 40 do
        point, parent, relative, x, y, width, height = unpack(tRA["TRA"..i]);
        ref = CreateFrame("Button", "DHUD3_TRA"..i, _G["DHUD3_"..parent]);
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y);
        ref:SetHeight(height);
        ref:SetWidth(width);
        local font = ref:CreateFontString("DHUD3_TRA"..i.."_Text", "OVERLAY");
        font:SetFontObject(GameFontHighlightSmall);
        font:SetText("");
        font:SetJustifyH("CENTER");
        font:SetWidth(15);
        font:SetFont("Fonts/FRIZQT__.TTF", 12, "OUTLINE");
        font:Show();
        font:ClearAllPoints();
        font:SetPoint("CENTER", ref, "CENTER", 0, 0);
        bgt = ref:CreateTexture("DHUD3_TRA"..i.."_Border", "ARTWORK");
        bgt:SetTexture("Interface\\Buttons\\UI-Debuff-Border");
        bgt:SetPoint("BOTTOM", ref, "BOTTOM", 0, 0);
        bgt:SetHeight(height);
        bgt:SetWidth(width);
        bgt:SetTexCoord(0, 1, 0, 1);
        ref:SetNormalTexture( "Interface\\Icons\\Ability_Druid_TravelForm" );
        ref:SetScript("OnEnter", function(this)
                if ( not this:IsVisible() ) then
                    return;
                end
                GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT");
                if this.hasDebuff then
                    GameTooltip:SetUnitDebuff("target", this.id);
                else
                    GameTooltip:SetUnitBuff("target", this.id);
                end
            end )
        ref:SetScript("OnLeave", function() 
                GameTooltip:Hide();
            end )        
        ref:EnableMouse(false);
        ref:Hide();
    end
    
    for i = 1, 40 do
        point, parent, relative, x, y, width, height = unpack(tLA["TLA"..i]);
        ref = CreateFrame("Button", "DHUD3_TLA"..i, _G["DHUD3_"..parent]);
        ref:ClearAllPoints();
        ref:SetPoint(point, _G["DHUD3_"..parent], relative, x, y);
        ref:SetHeight(height);
        ref:SetWidth(width);
        local font = ref:CreateFontString("DHUD3_TLA"..i.."_Text", "ARTWORK");
        font:SetFontObject(GameFontHighlightSmall);
        font:SetText("");
        font:SetJustifyH("RIGHT");
        font:SetJustifyV("BOTTOM");
        font:SetWidth(width+1);
        font:SetFont("Fonts/FRIZQT__.TTF", 12, "OUTLINE");
        font:Show();
        font:ClearAllPoints();
        font:SetPoint("CENTER", ref, "CENTER", 0, 0);
        bgt = ref:CreateTexture("DHUD3_TLA"..i.."_Border","OVERLAY");
        bgt:SetTexture("Interface\\Buttons\\UI-Debuff-Border");
        bgt:SetPoint("BOTTOM", ref, "BOTTOM", 0, 0);
        bgt:SetHeight(height);
        bgt:SetWidth(width);
        bgt:SetTexCoord(0, 1, 0, 1);
        ref:SetNormalTexture( "Interface\\Icons\\Ability_Druid_TravelForm" );
        ref:SetScript("OnEnter", function(this)
                if ( not this:IsVisible() ) then
                    return;
                end
                GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT");
                if this.hasDebuff then
                    GameTooltip:SetUnitDebuff("target", this.id);
                else
                    GameTooltip:SetUnitBuff("target", this.id);
                end
            end )
        ref:SetScript("OnLeave", function() 
                GameTooltip:Hide();
            end )        
        ref:EnableMouse(false);
        ref:Hide();
    end
    
    
    -- Player/Target Dropdown
    ref = CreateFrame("Button", "DHUD3_targetDropDown", _G[UIParent], "UIDropDownMenuTemplate");
    ref:ClearAllPoints();
    ref:SetPoint("TOP", _G[UIParent], "CENTER", 10, -60);
    ref:SetWidth(10);
    ref:SetHeight(10);    
    ref:Hide();
    
    ref = CreateFrame("Button", "DHUD3_playerDropDown", _G[UIParent], "UIDropDownMenuTemplate");
    ref:ClearAllPoints();
    ref:SetPoint("TOP", _G[UIParent], "CENTER", 10, -60);
    ref:SetWidth(10);
    ref:SetHeight(10);
    ref:SetScript("OnLoad", function(this) 
                UIDropDownMenu_Initialize(DHUD3_playerDropDown, PlayerDropDownInit, "MENU")
                end
            )
    ref:Hide();
    
    -- Configure Right click funtionality
    DHUD3_target_Text:RegisterForClicks('RightButtonUp');
    DHUD3_target_Text:SetAttribute("type2", "menu");
    local showmenu = function()
                ToggleDropDownMenu(1, nil, DHUD3_targetDropDown, "DHUD3_target_Text", 120, 10)
            end
    DHUD3_target_Text.menu = showmenu;
    
    -- Enable targettarget click to target
    _G["DHUD3_targetTarget_Text"]:SetAttribute("unit", "targetTarget");
    _G["DHUD3_targetTarget_Text"]:SetAttribute("type1", "target");
end

function Player:OnEnable()
    
	--DHUD3:Debug("Player:OnEnable");	
    PlaceBarTexts();
    self:Refresh();
	if not self:IsEnabled() then return end
	
	self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED");
    self:RegisterEvent("PLAYER_FLAGS_CHANGED");
    self:RegisterEvent("PLAYER_UPDATE_RESTING");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
    self:RegisterEvent("RAID_TARGET_UPDATE");
    self:RegisterEvent("UNIT_FACTION");
    UIDropDownMenu_Initialize(DHUD3_targetDropDown, TargetDropDownInit, "MENU")
end

function Player:OnDisable()
    
	self:SendMessage("DHUD3_TARGETBARS_HIDE");
    self:UnregisterAllEvents();
    DHUD3_li:Hide();
    DHUD3_lm:Hide();
    DHUD3_ri:Hide();
    DHUD3_rm:Hide();
    DHUD3_li:SetScript("OnUpdate", nil);
    DHUD3_lm:SetScript("OnUpdate", nil);
    DHUD3_ri:SetScript("OnUpdate", nil);
    DHUD3_rm:SetScript("OnUpdate", nil);
    DHUD3_lpvp:Hide();
	DHUD3_rpvp:Hide();
    DHUD3_lleader:Hide();
	DHUD3_llooter:Hide();
	DHUD3_lcombat:Hide();
	DHUD3_ltelite:Hide();
	DHUD3_rleader:Hide();
	DHUD3_rlooter:Hide();
	DHUD3_rcombat:Hide();
	DHUD3_rtelite:Hide();
	DHUD3_lrest:Hide();
    DHUD3_rrest:Hide();
end

function Player:Refresh()

	db = self.db.profile
	if not self:IsEnabled() then return end
    local maxVal, curFrac,needH, needP;
    local color = {r=0,g=0,b=0,hex=""};
    
    ConfigBars();
    DHUD3_li:Hide();
    DHUD3_lm:Hide();
    DHUD3_ri:Hide();
    DHUD3_rm:Hide();
    DHUD3_li:SetScript("OnUpdate", nil);
    DHUD3_lm:SetScript("OnUpdate", nil);
    DHUD3_ri:SetScript("OnUpdate", nil);
    DHUD3_rm:SetScript("OnUpdate", nil);
    
    -- Fonts
    local font
    local w
    --Targer Text
    local scale = DHUD3.GetScale();
    DHUD3_target_Text_Text:SetFont(defaultfont, db.tNSize * scale)  
    --Targer Target Text
    DHUD3_targetTarget_Text_Text:SetFont(defaultfont, db.totNSize * scale)
    -- Player Bar's Text
    font = _G["DHUD3_"..playerLayout.pHealth.."_Text_Text"]
    font:SetFont(defaultfont, db.pBarTextSize * scale)
    font:SetWidth(200)
    font = _G["DHUD3_"..playerLayout.pPower.."_Text_Text"]
    font:SetFont(defaultfont, db.pBarTextSize * scale)
    font:SetWidth(200)
    
    -- Target Bar's Text
    font = _G["DHUD3_"..playerLayout.tHealth.."_Text_Text"]
    font:SetFont(defaultfont, db.tBarTextSize * scale)
    font:SetWidth(200)
    font = _G["DHUD3_"..playerLayout.tPower.."_Text_Text"]
    font:SetFont(defaultfont, db.tBarTextSize * scale)
    font:SetWidth(200)
    
    -- Player
    Player:RegisterEvent("PLAYER_TARGET_CHANGED");
        -- Health
    maxVal, curFrac, needH = DHUD3:GetUnitHealth("player");
    local colors = db.pColors[tostring(7)];
    color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
    DHUD3:UpdateBar(playerLayout.pHealth, maxVal, curFrac, color);
    local bar = _G["DHUD3_"..playerLayout.pHealth]
    bar:UnregisterAllEvents();
    bar.who = "player";
    bar.bar = playerLayout.pHealth;
    bar.colors = colors;
    bar:RegisterEvent("UNIT_HEALTH");
    bar:RegisterEvent("UNIT_MAXHEALTH");
    bar:SetScript("OnEvent", OnUnitHealth);
    bar:Show();
        --Power
    maxVal, curFrac, needP = DHUD3:GetUnitPower("player");
    colors = db.pColors[tostring(UnitPowerType("player"))];
    color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
    DHUD3:UpdateBar(playerLayout.pPower, maxVal, curFrac, color);    
    bar = _G["DHUD3_"..playerLayout.pPower]
    bar:UnregisterAllEvents();
    bar.who = "player";
    bar.bar = playerLayout.pPower;
    bar.maxVal = maxVal;
    bar.curFrac = curFrac;
    local powerType = UnitPowerType("player");
    bar.display = powerType;
    bar.colors = colors;
    bar:SetScript("OnUpdate", OnUpdateBarPower);
    if ( powerType == 0 ) then
        bar:RegisterEvent("UNIT_MAXMANA");
    elseif ( powerType == 1 ) then
        bar:RegisterEvent("UNIT_MAXRAGE");
    elseif ( powerType == 3 ) then
        bar:RegisterEvent("UNIT_MAXENERGY");
    elseif ( powerType == 6 ) then
        bar:RegisterEvent("UNIT_MAXRUNIC_POWER");
    end
    bar:RegisterEvent("UNIT_DISPLAYPOWER");
    bar:SetScript("OnEvent", OnUnitPower);
    bar:Show();
    
        -- Status: Rest, PvP, Leader, Looter
    DHUD3_lrest:Hide();
    DHUD3_rrest:Hide();
	DHUD3_lpvp:Hide();
	DHUD3_rpvp:Hide();
    DHUD3_lleader:Hide();
	DHUD3_llooter:Hide();
	DHUD3_lcombat:Hide();
	DHUD3_ltelite:Hide();
	DHUD3_rleader:Hide();
	DHUD3_rlooter:Hide();
	DHUD3_rcombat:Hide();
	DHUD3_rtelite:Hide();
	
	if IsResting() and db.restIcon then
        _G["DHUD3_"..playerLayout.pIcons.."rest"]:Show();
    end
    UpdatePlayerPvP();
    
    if ( GetNumPartyMembers() > 0 and db.leaderIcon ) then
        if ( UnitIsPartyLeader("player") ) then
            _G["DHUD3_"..playerLayout.pIcons.."leader"]:Show();
            local lootmethod = GetLootMethod();
            if ( lootmethod == "master" ) then
                _G["DHUD3_"..playerLayout.pIcons.."looter"]:Show();
            end
        end
    end
    
    -- Target & Target of Target
	HideTargetAuras();
    if ( UnitExists("target") ) then
        self:PLAYER_TARGET_CHANGED();
    end
    
    --Target Auras Tips
    if not db.tSwap then
		buffFrame   = "DHUD3_TRA";
		debuffFrame = "DHUD3_TLA";
	else
		buffFrame   = "DHUD3_TLA";    
		debuffFrame = "DHUD3_TRA";
	end
    for i = 1, 40 do
		_G[buffFrame..i]:EnableMouse(db.tTips);
        _G[debuffFrame..i]:EnableMouse(db.tTips);
    end
    
    -- Target and TargetofTarget Tags
	if ( db.tName ) then
		local tagCode = (string_len(db.tCStyle) > 0) and db.tCStyle or db.tStyle;
		DogTag:AddFontString(_G["DHUD3_target_Text_Text"], _G["DHUD3_target_Text"], tagCode, "Unit", { unit = "target" } );
	else
		DogTag:RemoveFontString(_G["DHUD3_target_Text_Text"]);
		_G["DHUD3_target_Text"]:Hide();
	end
	if ( db.totName ) then
		local tagCode = (string_len(db.totCStyle) > 0) and db.totCStyle or db.totStyle;
		DogTag:AddFontString(_G["DHUD3_targetTarget_Text_Text"], _G["DHUD3_targetTarget_Text"], tagCode, "Unit", { unit = "targettarget" } );
	else
		DogTag:RemoveFontString(_G["DHUD3_targetTarget_Text_Text"]);
		_G["DHUD3_targetTarget_Text"]:Hide();
	end
    
    -- Pet Swap
    if db.swapPet then
        self:RegisterEvent("UNIT_ENTERED_VEHICLE");
        self:RegisterEvent("UNIT_EXITED_VEHICLE");
        if ( UnitInVehicle("player") ) then
            SwapVehicle(true);
        end
    else
        self:UnregisterEvent("UNIT_ENTERED_VEHICLE");
        self:UnregisterEvent("UNIT_EXITED_VEHICLE");
    end
end

function Player:MoveFrame()

	if ( IsAltKeyDown() ) then
		this:StartMoving();
		this.isMoving = true;
	end
end

function Player:DropFrame()
	
	if ( this.isMoving ) then
		this.isMoving = false;
		this:StopMovingOrSizing();
		this:SetUserPlaced(false);
		local barText = DHUD3.layout.barText;
		local i = string_find(this:GetName(), "_");
		bar = string.sub(this:GetName(), i+1, -6);
		local point, parent, relative, _, _, _, _, justify = unpack(barText[bar]);
		local pointOfs, relativeTo, relativePoint, xOfs, yOfs = this:GetPoint();
		--Calculate relative position according to frame creation
		if ( justify == "LEFT" ) then
			xp = xOfs + 28;
		else
			xp = xOfs - 28;
		end
		yp = yOfs + 135;
		-- Save button position to profile
		db.playerBarsText[bar].moved = true;
		db.playerBarsText[bar].x = xp;
		db.playerBarsText[bar].y = yp;
		this:ClearAllPoints();
		this:SetPoint(point, _G["DHUD3_"..parent], relative, xp, yp);
		pointOfs, relativeTo, relativePoint, xOfs, yOfs = this:GetPoint();
	end
end

function Player:PLAYER_TARGET_CHANGED()

    ----self:Debug("PLAYER_TARGET_CHANGED")
    local maxVal, curFrac, typ;
    local color = {r=0,g=0,b=0,hex=""};
    
    if ( UnitExists("target") and db.tBars and not ( DHUD3:UnitIsNPC("target") and not db.tNPC ) ) then
        local healthTyp, powerTyp = 7, UnitPowerType("target");
		if ( UnitIsTapped("target") and ( not UnitIsTappedByPlayer("target") ) ) then       
            healthTyp = 8;
            powerTyp = 8;
		end
        --Health
        maxVal, curFrac, needH = DHUD3:GetUnitHealth("target");
        local colors = db.tColors[tostring(healthTyp)];
        color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
        DHUD3:UpdateBar(playerLayout.tHealth, maxVal, curFrac, color);
        local bar = _G["DHUD3_"..playerLayout.tHealth];
        bar.who = "target";
        bar.bar = playerLayout.tHealth;
        bar.colors = colors;
        bar:UnregisterAllEvents();
        bar:RegisterEvent("UNIT_HEALTH");
        bar:RegisterEvent("UNIT_MAXHEALTH");
        bar:SetScript("OnEvent", OnUnitHealth);
        bar:Show();
        --Power
        maxVal, curFrac, needP = DHUD3:GetUnitPower("target");
        colors = db.tColors[tostring(powerTyp)];
        color.r, color.g, color.b, color.hex = DHUD3:Colorize(colors, curFrac);
        DHUD3:UpdateBar(playerLayout.tPower, maxVal, curFrac, color);
        bar = _G["DHUD3_"..playerLayout.tPower];
        bar.who = "target";
        bar.bar = playerLayout.tPower;
        bar.maxVal = maxVal;
        bar.curFrac = curFrac;
        bar.display = powerTyp;
        bar.colors = colors;
        if ( maxVal ~= 0 ) then 
            bar:SetScript("OnUpdate", OnUpdateBarPower);
            bar:UnregisterAllEvents();
            if ( powerTypw ~= 8 ) then
                if ( powerType == 0 ) then
                    bar:RegisterEvent("UNIT_MAXMANA");
                elseif ( powerType == 1 ) then
                    bar:RegisterEvent("UNIT_MAXRAGE");
                elseif ( powerType == 3 ) then
                    bar:RegisterEvent("UNIT_MAXENERGY");
                elseif ( powerType == 6 ) then
                    bar:RegisterEvent("UNIT_MAXRUNIC_POWER");
                end
                bar:RegisterEvent("UNIT_DISPLAYPOWER");
                bar:SetScript("OnEvent", OnUnitPower);
            end
        else
            bar:SetScript("OnUpdate", nil);
            bar:UnregisterAllEvents();
        end
        bar:Show();
        if ( db.tName ) then
            _G["DHUD3_target_Text"]:EnableMouse(true);
            _G["DHUD3_target_Text"]:Show()
			self:RegisterEvent("UNIT_TARGET");
            if ( UnitExists("targetTarget") and db.totName ) then
                _G["DHUD3_targetTarget_Text"]:Show();
            else
                _G["DHUD3_targetTarget_Text"]:Hide();
            end
        else
            _G["DHUD3_target_Text"]:EnableMouse(false);
            _G["DHUD3_target_Text"]:Hide()
        end
		if ( db.tAuras ) then 
			UpdateTargetAuras();
			self:RegisterEvent("UNIT_AURA");
		end
        UpdateTargetElite(true);
        DHUD3:UpdateAlpha(2, true);
    else
        _G["DHUD3_"..playerLayout.tPower]:Hide();
        _G["DHUD3_"..playerLayout.tHealth]:Hide();
        _G["DHUD3_target_Text"]:Hide()
        _G["DHUD3_targetTarget_Text"]:Hide();
        UpdateTargetElite(false);
        self:UnregisterEvent("UNIT_AURA");
        HideTargetAuras();
        DHUD3:UpdateAlpha(2, false);
    end
    self:RAID_TARGET_UPDATE();
end

function Player:UNIT_TARGET(event, who)
	
	--DHUD3:Debug("Player:UNIT_TARGET", who);
	if ( who == "target" ) then
		if ( UnitExists("targetTarget") and db.totName ) then
			_G["DHUD3_targetTarget_Text"]:Show();
		else
			_G["DHUD3_targetTarget_Text"]:Hide();
		end
	end
end

function Player:PLAYER_FLAGS_CHANGED()

    --DHUD3:Debug("PLAYER_FLAGS_CHANGED", UnitIsPartyLeader("player"), GetPartyLeaderIndex()) 
    if ( GetNumPartyMembers() > 0 ) then 
        if ( GetPartyLeaderIndex() ~= 0 and db.leaderIcon ) then
            _G["DHUD3_"..playerLayout.pIcons.."leader"]:Show();
        else
            _G["DHUD3_"..playerLayout.pIcons.."leader"]:Hide();
        end
    else
        _G["DHUD3_"..playerLayout.pIcons.."leader"]:Hide();
        _G["DHUD3_"..playerLayout.pIcons.."looter"]:Hide();
    end
end

function Player:RAID_TARGET_UPDATE()
    
    if ( db.tName and db.tPvpIcon ) then
        local _, _, _, _, _, _, _, x0, x1, y0, y1, path = unpack(DHUD3.layout.icons.target);
        if ( UnitIsPVPFreeForAll("target")  ) then
            path = path[3];
        elseif ( UnitIsPVP("target") ) then 
            local faction = UnitFactionGroup("target");
            if ( faction == "Horde" ) then
                path = path[2];   
            else
                path = path[1];               
            end
        else
            local index = GetRaidTargetIndex("target");
            if ( index ) then
                if ( index > 0 and index <= 8 ) then
                    path, x0, x1, y0, y1 = unpack( DHUD3.layout.raidIcons[index] );
                end
            end
        end
        _G["DHUD3_target_Texture"]:SetTexture(path);
        _G["DHUD3_target_Texture"]:SetTexCoord(x0,x1,y0,y1);
        _G["DHUD3_target"]:Show();
    else
        _G["DHUD3_target"]:Hide();
    end
end

function Player:PARTY_LOOT_METHOD_CHANGED()

    --DHUD3:Debug("PARTY_LOOT_METHOD_CHANGED")
    local _, master = GetLootMethod();
    if ( master == 0 and db.leaderIcon ) then
        _G["DHUD3_"..playerLayout.pIcons.."looter"]:Show();
    else
        _G["DHUD3_"..playerLayout.pIcons.."looter"]:Hide();
    end
end

function Player:PLAYER_UPDATE_RESTING()

    ----self:Debug("PLAYER_UPDATE_RESTING")
    if ( IsResting() and db.restIcon ) then
        _G["DHUD3_"..playerLayout.pIcons.."rest"]:Show();
    else
        _G["DHUD3_"..playerLayout.pIcons.."rest"]:Hide();
    end
end

function Player:UNIT_AURA(event, who)

	----self:Debug("UNIT_AURA", who)
    if ( who == "target" ) then
	    UpdateTargetAuras();
    end
end

function Player:UNIT_FACTION(event, who)

    if ( who == "player" ) then
        UpdatePlayerPvP();
    end
end

function Player:UNIT_ENTERED_VEHICLE(event, who)
	
	--DHUD3:Debug("UNIT_ENTERED_VEHICLE");
    if who == "player" then
        if ( UnitInVehicle("player") ) then
            SwapVehicle(true);
        end
    end
end

function Player:UNIT_EXITED_VEHICLE(event, who)
	
	--DHUD3:Debug("UNIT_EXITED_VEHICLE");
    if ( who == "player" and _G["DHUD3_"..playerLayout.pHealth].swap ) then
        SwapVehicle(false);
    end
end

function Player:PLAYER_REGEN_ENABLED()
    
	--self:Debug("PLAYER_REGEN_ENABLED")
    _G["DHUD3_"..playerLayout.pIcons.."combat"]:Hide()
end

function Player:PLAYER_REGEN_DISABLED()
    
	--self:Debug("PLAYER_REGEN_DISABLED")
    if db.combatIcon then
        _G["DHUD3_"..playerLayout.pIcons.."combat"]:Show()
    end
end  
